#include <genieArduino.h>
#include <SPI.h>
#include <SD.h>

#define  CS_LINE    10
SdFile   root;
SdFile   ipFold;
SdFile   file;
Sd2Card  card;
SdVolume volume;

#define  RESETLINE  4

#define  GENIE_OBJ_MAGIC 34

#define  NO_REPLY   -1
#define  ERR        0
#define  OK         1
#define  COMPLETE   2

#define  USD_FAILED       -1
#define  GENIE_NACK       0
#define  GENIE_FAILED     1
#define  GENIE_COMPLETED  2

Genie genie;

#define MAX_SIZE    510

const char *reply[3] = {"ERROR", "OK ", "COMPLETED"};

bool updateDisplay    = false;
bool waitingForReply  = false;
uint8_t magicReply;


void setup() {
  pinMode(13, OUTPUT); // indicator led

  genie.AttachEventHandler(myGenieEventHandler);
  genie.AttachMagicByteReader(myGenieMagicByteReader);
  genie.AttachMagicDoubleByteReader(myGenieMagicDoubleByteReader);

  Serial.begin(115200);  // UART Debug for terminal
  Serial1.begin(115200); // UART for Genie Display

  genie.Begin(Serial1);
  genie.debug(Serial, 6);

  if (!initSDcard(CS_LINE, SPI_FULL_SPEED)) {
    return;
  }

}

bool initSDcard(int csLine, int SPI_SPEED) {

  pinMode(csLine, OUTPUT);
  Serial.println("Initializing SD card...");
  if (!SD.begin(csLine)) {
    Serial.println("Initialization failed!");
    return false;
  }

  // Init code for Unix style access ('simply' so we can get date/time)
  //
  // If SPI_FULL_SPEED doesn't work with your setup, try SPI_HALF_SPEED
  //
  if (!card.init(SPI_SPEED, csLine)) {
    Serial.println("Initialization failed!");
    return false;
  }

  // Initialize a FAT volume
  if (!volume.init(&card)) {
    Serial.println("Initialization failed!");
    return false;
  }
  Serial.print("Volume is FAT");
  Serial.println(volume.fatType(), DEC);
  Serial.println();

  // open the root directory
  if (!root.openRoot(&volume) || !ipFold.openRoot(&volume))
  {
    Serial.println("Initialization failed!");
    return false;
  }
  // End Init code for Unix style access

  Serial.println("Initialization Completed.");
  return true;

}

int update4xe(char *directory, uint8_t bank = 1) {
  if (bank > 5 || !bank) return false;
  gotoDir(directory);

  char *bankFile = "RunBankN.4xe";
  uint16_t versionDetails[4];

  bankFile[7] = bank + '0';

  if (!getDetails(bankFile, versionDetails)) {
    return USD_FAILED;
  }
  versionDetails[3] = bank;

  if (genie.WriteMagicBytes(0, (uint8_t*) versionDetails, 8) != 1) {
    Serial.println("Failed to set parameters");
    return GENIE_NACK;
  }

  int packets = versionDetails[0] / MAX_SIZE;
  int lastPacketSize = versionDetails[0] % MAX_SIZE;
  if (!file.open(&ipFold, bankFile, O_READ)) {
    Serial.println("Failed to open file");
    return USD_FAILED;
  }
  Serial.println("File Ready to be Copied");

  genie.timeout(5000); // Increase if necessary

  for (int i = 0; i < packets; i++) {
    Serial.print("Packets Remaining: "); Serial.println(packets - i + ((lastPacketSize) ? 1 : 0));
    uint8_t data[MAX_SIZE] = {0};
    file.read(data, MAX_SIZE);
    do {
      magicReply = NO_REPLY;
      waitingForReply = true;

      uint8_t result = genie.WriteMagicDBytes(0, (uint16_t*)data, 255, GENIE_REPORT_OBJ); // Wait for GENIE_REPORT_OBJ reply
      if (result == -1) {
        Serial.println("Got timeout. Consider increasing timeout using genie.timeout()");
        file.close();
        return GENIE_FAILED;
      }
      genie.DoEvents(); // Process GENIE_REPORT_OBJ message

      if (waitingForReply) { // If the GENIE_REPORT_OBJ is not from MagicObject0
        Serial.println("Reply wasn't from MagicObject0.");
        Serial.println("Try stopping all ReadObject functions and clearing the RX Buffer before executing update4xe function.");
        file.close();
        return GENIE_FAILED;
      }

      if (magicReply == ERR) {
        Serial.println("Display sent error message");
      }

    } while (magicReply == ERR);
  }

  if (lastPacketSize) {
    Serial.print("Packets Remaining: "); Serial.println(1);
    uint8_t data[lastPacketSize];
    memset(data, 0, lastPacketSize);
    file.read(data, lastPacketSize);
    do {
      magicReply = NO_REPLY;
      waitingForReply = true;

      uint8_t result = genie.WriteMagicDBytes(0, (uint16_t*)data, (uint8_t)(lastPacketSize / 2 + lastPacketSize % 2), GENIE_REPORT_OBJ); // Wait for GENIE_REPORT_OBJ reply

      if (result == -1) {
        Serial.println("Got timeout. Consider increasing timeout using genie.timeout()");
        file.close();
        return GENIE_FAILED;
      }
      genie.DoEvents(); // Process GENIE_REPORT_OBJ message

      if (waitingForReply) { // If the GENIE_REPORT_OBJ is not from MagicObject0
        Serial.println("Reply wasn't from MagicObject0.");
        Serial.println("Try stopping all ReadObject functions and clearing the RX Buffer before executing update4xe function.");
        file.close();
        return GENIE_FAILED;
      }

      if (magicReply == ERR) {
        Serial.println("Display sent error message");
      }

    } while (magicReply == ERR);
  }
  file.close();

  if (magicReply != COMPLETE) {
    Serial.println("Display didn't send complete message");
    return GENIE_FAILED;
  }

  resetDisplay();
  return GENIE_COMPLETED;

}


bool gotoRoot() {
  ipFold.close();
  if (!ipFold.openRoot(&volume))
  {
    Serial.println("Open Root Failed");
    return false;
  }
  Serial.println("Open Root Successful");
  return true;
}

bool openFolder(String folder) {
  char _folder[20];
  folder.toCharArray(_folder, folder.length() + 1);

  SdFile temp = ipFold;
  ipFold.close();
  if (!ipFold.open(&temp, _folder, O_READ))
  {
    Serial.println("Open " + (String)_folder + " Failed");
    return false;
  }
  Serial.println("Open " + (String)_folder + " Successful");
  return true;
}

bool gotoDir(char* directory) {
  String dir;
  if (directory[0] == '/') {
    Serial.print("Directory : "); Serial.println(directory);
    if (!gotoRoot()) return false; // Start at root folder
    Serial.print("Folder    : "); Serial.println("/");
    dir = (String)(&directory[1]);
  } else {
    dir = (String)directory;
  }
  while (dir.length() > 0) {
    Serial.print("Directory : "); Serial.println(dir);
    int idx = dir.indexOf('/');
    if (idx > 0) {
      String folder = dir.substring(0, idx);
      Serial.print("Folder    : "); Serial.println(folder);
      if (!openFolder(folder)) return false;
      dir = dir.substring(idx + 1);
    } else {
      Serial.print("Folder    : "); Serial.println(dir);
      if (openFolder(dir)) return false;
      break;
    }
  }
  return true;
}

bool getDetails(char* filename, uint16_t details[3]) {
  Serial.print("Getting details of "); Serial.println(filename);
  if (file.open(&ipFold, filename, O_READ)) {
    dir_t d;
    file.dirEntry(&d);
    details[0] = file.fileSize();
    details[1] = d.lastWriteDate;
    details[2] = d.lastWriteTime;
    file.close();
    Serial.println("Got Details...");
    return true;
  }
  Serial.println("Failed to get details");
  return false;
}

void myGenieEventHandler() {
  genieFrame Event;
  genie.DequeueEvent(&Event);

  if (Event.reportObject.cmd == GENIE_REPORT_OBJ) {
    if (Event.reportObject.object == GENIE_OBJ_MAGIC) {
      if (Event.reportObject.index == 0) {
        if (waitingForReply) {
          waitingForReply = false;
          magicReply = genie.GetEventData(&Event);
          if (magicReply <= COMPLETE) {
            Serial.print("Received "); Serial.println(reply[magicReply]);
          }
        }
      }
    }
  } else if (Event.reportObject.cmd == GENIE_REPORT_EVENT) {
    if (Event.reportObject.object == GENIE_OBJ_WINBUTTON) {
      if (Event.reportObject.index == 0) {
        updateDisplay = true;
      }
    }
  } else if (Event.reportObject.cmd == GENIE_PING) {

    if (Event.reportObject.object == GENIE_DISCONNECTED) {
      digitalWrite(13, HIGH);
      Serial.println("Display got disconnected");
    } else if (Event.reportObject.object == GENIE_READY) {
      digitalWrite(13, LOW);
      Serial.println("Display reconnected");
    } else if (Event.reportObject.object == GENIE_ACK) {
      // manual ping response - Display is Connected - Do something if required
    } else if (Event.reportObject.object == GENIE_NAK) {
      // manually ping response - Display is NOT Connected - Do something if required
    }

  }
}

void myGenieMagicByteReader(uint8_t index, uint8_t len) {
  for (uint8_t i; i < len; i++)
    (void)genie.GetNextByte();
  (void)genie.GetNextByte(); // Checksum
}

void myGenieMagicDoubleByteReader(uint8_t index, uint8_t len) {
  for (uint8_t i; i < len; i++)
    (void)genie.GetNextDoubleByte();
  (void)genie.GetNextByte(); // Checksum
}

void loop() {
  static long waitPeriod = millis(); // timer to repeat task
  genie.DoEvents();

  if (millis() >= waitPeriod) {
    if (genie.online()) {
      // Do something:
      if (updateDisplay) {
        int res = update4xe("/Updates");
        switch (res) {
          case USD_FAILED:
          case GENIE_NACK:
          case GENIE_FAILED:
            Serial.println("Retrying...");
            // if (GENIE_FAILED) resetDisplay();
            break;
          case GENIE_COMPLETED:
            updateDisplay = false;
            Serial.println("Update Successful");
            break;
          default:
            break;
        }
      }
    }
    waitPeriod = millis() + 10;
  }
}

void resetDisplay() {

  pinMode(RESETLINE, OUTPUT);  // Set D4 on Arduino to Output (4D Arduino Adaptor V2 - Display Reset)
  digitalWrite(RESETLINE, 1);  // Reset the Display via D4
  unsigned long startTime = millis();
  
  Serial.println("Resetting...");
  
  while (genie.online()) genie.DoEvents();

  genie.DoEvents();            // Execute another DoEvents to process Disconnect Flag
  Serial.print("Disconnected after "); Serial.print(millis() - startTime); Serial.println("ms");
  
  digitalWrite(RESETLINE, 0);  // unReset the Display via D4

  Serial.println("\nWaiting for reconnect");
  genie.timeout(1250); // Return timeout to default
  genie.recover(500);

  while (!genie.online()) {
    genie.DoEvents(); // Wait until display is ready
  }

  genie.recover(50);
  Serial.println("\nDisplay is ready");
}

